use std::collections::HashMap;

use chrono::Timelike;
use serde::{Deserialize, Serialize};

use crate::{
    clock,
    forecast::{Forecast, ForecastSet, Rate},
};

const DATA: &'static str = include_str!("../data.json");

#[derive(Serialize, Deserialize, Debug)]
pub struct Data {
    pub db_data: SubData,
    pub fish_entries: Vec<FishMeta>,
    #[serde(skip)]
    pub forecasts: ForecastSet,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct SubData {
    #[serde(alias = "FISH")]
    pub fish: HashMap<u32, FishEntry>,
    #[serde(alias = "ZONES")]
    pub zones: HashMap<u32, Zone>,
    #[serde(alias = "WEATHER_RATES")]
    pub weather_rates: HashMap<u32, WeatherRate>,
    #[serde(alias = "FISHING_SPOTS")]
    pub fishing_spots: HashMap<u32, FishingSpot>,
    #[serde(alias = "WEATHER_TYPES")]
    pub weather_types: HashMap<u32, WeatherType>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(deserialize = "camelCase"))]
pub struct FishEntry {
    #[serde(alias = "_id")]
    pub id: u32,
    pub start_hour: Option<f32>,
    pub end_hour: Option<f32>,
    pub location: Option<u32>,
    pub best_catch_path: Vec<u32>,
    pub predators: Vec<Vec<u32>>,
    pub intuition_length: Option<u32>,
    pub patch: f32,
    pub folklore: Option<u32>,
    pub fish_eyes: bool,
    pub big_fish: bool,
    pub weather_set: Vec<u32>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Zone {
    pub name_en: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct FishingSpot {
    #[serde(alias = "_id")]
    pub id: u32,
    pub name_en: String,
    #[serde(alias = "territory_id")] // not sure why it's named this, but thanks serde!
    pub forecast_id: u32,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct WeatherRate {
    pub zone_id: u32,
    pub weather_rates: Vec<Vec<u32>>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct WeatherType {
    pub name_en: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Location {
    pub id: u32,
    pub name_en: String,
    pub map_coords: (f32, f32, u32),
}

#[derive(Serialize, Deserialize, Debug)]
pub struct FishMeta {
    pub id: u32,
    pub name_en: String,
}

pub struct CombinedFish<'a> {
    pub entry: &'a FishEntry,
    pub meta: &'a FishMeta,
}

impl<'a> CombinedFish<'a> {
    pub fn is_in_time_range(&self) -> bool {
        if self.entry.start_hour.is_none() || self.entry.end_hour.is_none() {
            return false;
        }
        let et = clock::get_current_eorzea_date();
        let start_hour = self.entry.start_hour.unwrap();
        let end_hour = self.entry.end_hour.unwrap();
        let spans_midnight = start_hour > end_hour;
        let cur_hour = et.hour() as f32 + et.minute() as f32 / 60.;
        if spans_midnight {
            return cur_hour > start_hour || cur_hour < end_hour;
        } else {
            return start_hour < cur_hour && cur_hour < end_hour;
        }
    }

    pub fn get_forecast<'b>(&self, data: &'b Data) -> Option<&'b Forecast> {
        if self.entry.location.is_none() {
            return None;
        }
        // Find the associated fishing spot.
        let spot = data
            .db_data
            .fishing_spots
            .get(&self.entry.location.unwrap())
            .unwrap();
        data.forecasts.get(&spot.forecast_id)
    }

    pub fn is_in_correct_weather(&self, forecast: &Forecast) -> bool {
        if self.entry.weather_set.is_empty() {
            return true;
        }
        let cur_weather = forecast.weather_now();
        self.entry
            .weather_set
            .iter()
            .any(|ws| ws == &cur_weather.weather_id)
    }

    pub fn is_up(&self, data: &Data) -> bool {
        let forecast = self.get_forecast(data);
        self.is_in_time_range()
            && forecast.is_none_or(|forecast| self.is_in_correct_weather(forecast))
    }
}

impl Data {
    pub fn new() -> Self {
        let json = &mut serde_json::Deserializer::from_str(DATA);
        let mut data: Self = serde_path_to_error::deserialize(json).unwrap();
        data.forecasts = data
            .db_data
            .weather_rates
            .iter()
            .map(|(id, wr)| {
                (
                    id.clone(),
                    Forecast {
                        zone_id: wr.zone_id,
                        rates: wr
                            .weather_rates
                            .iter()
                            .map(|rate| Rate {
                                weather_id: rate[0],
                                rate: rate[1],
                            })
                            .collect(),
                    },
                )
            })
            .collect();
        data
    }

    pub fn fish_with_meta(&self) -> HashMap<u32, CombinedFish> {
        self.db_data
            .fish
            .iter()
            .filter_map(|(k, v)| {
                let corresponding_meta = self.fish_entries.iter().find(|m| &m.id == k);
                corresponding_meta.map(|m| (k.clone(), CombinedFish { entry: v, meta: m }))
            })
            .collect()
    }
}

pub fn get_weather_name(data: &Data, id: u32) -> &str {
    &data.db_data.weather_types.get(&id).unwrap().name_en
}

pub fn get_zone_name(data: &Data, id: u32) -> &str {
    &data.db_data.zones.get(&id).unwrap().name_en
}
