use std::sync::Arc;

use axum::{extract::State, http::StatusCode, response::IntoResponse, routing::get, Router};
use data::{CombinedFish, Data};
use maud::{html, Markup};

pub mod clock;
pub mod data;
pub mod forecast;

pub struct AppState {
    pub data: Data,
}

pub struct AppError(anyhow::Error);

impl IntoResponse for AppError {
    fn into_response(self) -> axum::response::Response {
        (
            StatusCode::INTERNAL_SERVER_ERROR,
            format!("Something went wrong: {}", self.0),
        )
            .into_response()
    }
}

impl<E> From<E> for AppError
where
    E: Into<anyhow::Error>,
{
    fn from(err: E) -> Self {
        Self(err.into())
    }
}

#[axum::debug_handler]
async fn main_handler(state: State<Arc<AppState>>) -> Result<Markup, AppError> {
    let meta = state.data.fish_with_meta();
    let mut values: Vec<(&CombinedFish, bool)> = meta
        .values()
        .filter_map(|fish| {
            if fish.entry.big_fish {
                let is_up = fish.is_up(&state.data);
                Some((fish, is_up))
            } else {
                None
            }
        })
        .collect();
    values.sort_by(|(afish, aup), (bfish, bup)| bup.cmp(aup));
    Ok(html! {
        h1 { "Hello! Current ET " (clock::get_current_eorzea_date().format("%H:%M")) }
        @for (fish, is_up) in values {
            li {
                @if is_up {
                    "Up! "
                }
                (fish.meta.name_en)
                details {
                    @if fish.entry.start_hour.is_some() && fish.entry.end_hour.is_some() {
                        "From " (fish.entry.start_hour.unwrap()) "h to " (fish.entry.end_hour.unwrap()) "h"
                    }
                    @if fish.entry.weather_set.len() > 0 {
                        " Weather(s) " (fish.entry.weather_set.iter().map(|i| i.to_string()).collect::<Vec<_>>().join(", "))
                    }
                    @if let Some(forecast) = fish.get_forecast(&state.data) {
                        "Current weather in " (data::get_zone_name(&state.data, forecast.zone_id)) ": " (data::get_weather_name(&state.data, forecast.weather_now().weather_id))
                    }
                }
            }
        }
    })
}

#[tokio::main]
async fn main() {
    let app = Router::new()
        .route("/", get(main_handler))
        .with_state(Arc::new(AppState { data: Data::new() }));
    let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await.unwrap();
    println!("Listening on http://localhost:3000!");
    axum::serve(listener, app).await.unwrap();
}
